//  debyer -- program for calculation of diffration patterns
//  Copyright (C) 2006-2007 Marcin Wojdyr
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  $Id: fileio.cc 75 2008-07-20 15:18:03Z wojdyr $

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <cstring>
#include <cctype>
#include <cassert>
#include <map>
#include <vector>
#include <algorithm>
#include "atominfo.h"
#include "fileio.h"

using namespace std;


#define mcerr \
    if (dbr_nid == 0) \
        cerr 

int dbr_f_store_relative_coords = 0;

bool is_xyz_format(char const* buffer)
{
    // find new lines
    char *e[5];
    for (int i = 0; i < 5; ++i) {
        char const* start = i > 0 ? e[i-1]+1 : buffer;
        e[i] = strchr(start, '\n');
        if (!e[i])
            return false;
    }
    char* endptr = 0;
    strtol(buffer, &endptr, 10);
    while (isspace(*endptr))
        ++endptr;
    if (endptr < e[0])
        return false;
    for (int i = 2; i < 5; ++i) {
        char *ptr = e[i-1];
        while (isspace(*ptr))
            ++ptr;
        if (!isalpha(*ptr)) //should it start with letter?
            return false;
        while (isalnum(*ptr)) //atom name
            ++ptr;
        strtod(ptr, &endptr); //x
        ptr = endptr;
        strtod(ptr, &endptr); //y
        ptr = endptr;
        strtod(ptr, &endptr); //z
        ptr = endptr;
        if (ptr > e[i])
            return false;
        while (isspace(*ptr))
            ++ptr;
        if (ptr < e[i])
            return false;
    }
    return true;
}

bool is_plain_format(char const* /*buffer*/)
{
    return true;  //TODO
}

void read_xyz(InputData& in, dbr_aconf *aconf)
{
    const char* line = in.get_line(); //first line
    aconf->n = strtol(line, 0, 10);

    line = in.get_line(); //2nd line
    if (dbr_verbosity >= 0) 
        mcerr << aconf->n << " atoms, title: " << line << endl;

    aconf->atoms = new xyz_name[aconf->n];
    for (int i = 0; i < aconf->n; ++i) {
        xyz_name& atom = aconf->atoms[i];
        line = in.get_line(); //2nd line
        if (!line) {
            mcerr << "Error: Reading line " << i+3 << " failed." << endl;
            dbr_abort(EXIT_FAILURE);
        }
        if (sscanf(line, "%7s "DBR_F" "DBR_F" "DBR_F, atom.name, &atom.xyz[0], 
                    &atom.xyz[1], &atom.xyz[2]) != 4) {
            mcerr << "Format error in line " << i+3 << ":" << endl
                  << line << endl;
            dbr_abort(EXIT_FAILURE);
        }
    }
}


#define ASSERT_FORMAT(condition) \
    if (!(condition)) { \
        mcerr << "Error in line: " << in.get_line_number() << endl; \
        dbr_abort(EXIT_FAILURE); \
    }


static
void resize_atoms(size_t& atoms_size, xyz_name** atoms)
{
    xyz_name *new_atoms = new xyz_name[atoms_size*2];
    for (size_t i = 0; i < atoms_size; ++i)
        new_atoms[i] = (*atoms)[i];
    atoms_size *= 2;
    delete [] *atoms;
    *atoms = new_atoms;
}

void read_plain(InputData& in, dbr_aconf *aconf)
{
    size_t atoms_size = 1024;

    size_t counter = 0;
    aconf->atoms = new xyz_name[atoms_size];
    const char *line;
    while ((line = in.get_line())) {
        if (counter == atoms_size) 
            resize_atoms(atoms_size, &aconf->atoms);
        xyz_name& atom = aconf->atoms[counter];
        ++counter;
        const char *ptr = line;
        while (isspace(*ptr))
            ++ptr;
        string name;
        int r = 0;
        if (isalpha(*ptr)) { //name, x, y, z
            r = sscanf(line, "%7s "DBR_F" "DBR_F" "DBR_F, 
                        atom.name, &atom.xyz[0], &atom.xyz[1], &atom.xyz[2]);
        }
        else if (*ptr) {//x, y, z, name
            r = sscanf(line, DBR_F" "DBR_F" "DBR_F" %7s", 
                        &atom.xyz[0], &atom.xyz[1], &atom.xyz[2], atom.name);
        }
        if (!r) {
            mcerr << "Error in line " << counter << endl;
            dbr_abort(EXIT_FAILURE);
        }
    }
    aconf->n = counter;
}

void write_atoms_to_xyz_file(dbr_aconf const& aconf, string const& filename)
{
    if (dbr_nid != 0)
        return;
    if (dbr_verbosity >= 0)
        mcerr << "Writing atoms to XMOL xyz file: " << filename << endl;
    ofstream f(filename.c_str(), ios::out|ios::binary);
    if (!f) {
        mcerr << "Error. Can't open file " << filename << "for writing.\n";
        return;
    }
    f << aconf.n << "\nconverted by debyer\n";
    for (int i = 0; i < aconf.n; ++i) {
        xyz_name const& atom = aconf.atoms[i];
        f << atom.name << " " 
          << atom.xyz[0] << " " << atom.xyz[1] << " " << atom.xyz[2] << endl; 
    }
    f.close();
}

static bool endswith(string const& s, char const* e)
{
    int slen = s.size();
    int elen = strlen(e);
    return slen >= elen && strcmp(s.c_str() + slen - elen, e) == 0;
}

dbr_aconf read_atoms_from_file(InputData &in)
{
    dbr_aconf aconf;
    // initialize dbr_aconf
    aconf.n = 0;
    aconf.atoms = NULL;
    aconf.pbc.v00 = aconf.pbc.v01 = aconf.pbc.v02 = 0.;
    aconf.pbc.v10 = aconf.pbc.v11 = aconf.pbc.v12 = 0.;
    aconf.pbc.v20 = aconf.pbc.v21 = aconf.pbc.v22 = 0.;

    char const* cfg_magic = "Number of particles =";
    if (!strncmp(in.get_buffer(), cfg_magic, strlen(cfg_magic))) {
        if (dbr_verbosity >= 0)
            mcerr << "Detected AtomEye CFG format" << endl;
        read_atomeye(in, &aconf);
    }
    else if (endswith(in.get_filename(), "CONFIG") 
             || endswith(in.get_filename(), "REVCON")) {
        if (dbr_verbosity >= 0)
            mcerr << "DL_POLY CONFIG format" << endl;
        read_dlpoly_config(in, &aconf);
    }
    else if (is_xyz_format(in.get_buffer())) {
        if (dbr_verbosity >= 0)
            mcerr << "Detected XMol XYZ format" << endl;
        read_xyz(in, &aconf);
    }
    else if (is_plain_format(in.get_buffer())) {
        if (dbr_verbosity >= 0)
            mcerr << "Detected plain format with atom coordinates" << endl;
        read_plain(in, &aconf);
    }
    else {
        mcerr << "Error. Unknown format of input file" << endl;
        dbr_abort(EXIT_FAILURE);
    }
    if (dbr_verbosity > 0)
        mcerr << "Elapsed " << dbr_get_elapsed() << " s. Atoms were read." 
            << endl;
    aconf.orig_filename = in.get_filename();
    return aconf;
}

//=============   AtomEye file reading and writing    ==================

void read_atomeye(InputData& in, dbr_aconf *aconf)
{
    // only "A = 1.0 [Angstrom]" files are supported
    char const* cfg_magic = "Number of particles =";
    char const* extended_mark = "entry_count";
    int const magic_len = strlen(cfg_magic);
    double H[3][3] = {{0, 0, 0}, {0, 0, 0}, {0, 0, 0}};
    const char* line = in.get_line();
    assert(!strncmp(line, cfg_magic, magic_len));
    sscanf(line + magic_len, "%i", &aconf->n);
    if (dbr_verbosity >= 0) 
        mcerr << aconf->n << " atoms." << endl;
    aconf->atoms = new xyz_name[aconf->n];
    int counter = 0;
    bool extended = false;
    char ext_name[8] = {0, 0, 0, 0, 0, 0, 0, 0};

    while ((line = in.get_line())) {
        char const* nonblank = line;
        while (isspace(*nonblank))
            ++nonblank;

        // PBC line, H0(...) = ...
        if (nonblank[0] == 'H' && nonblank[1] == '0') {     
            char const* t = strchr(line, '('); 
            char *end;
            int a1 = strtol(++t, &end, 10);
            assert (a1 >= 1 && a1 <= 3);

            t = strchr(end, ','); 
            int a2 = strtol(++t, &end, 10);
            assert (a2 >= 1 && a2 <= 3);
            t = strchr(end, '='); 
            H[a1-1][a2-1] = strtod(++t, 0);
        }

        // entry_count line
        else if (!strncmp(nonblank, extended_mark, strlen(extended_mark))) {
            char *eq = strchr(nonblank, '=');
            ASSERT_FORMAT(*eq);
            int n;
            int r = sscanf(eq+1, "%i", &n);
            ASSERT_FORMAT(r == 1);
            if (dbr_verbosity > 0) 
                mcerr << "Extended CFG format with " << n 
                    << " entries per atom." << endl;
            extended = true;
        }

        // either atom mass or coordinates line
        else if (nonblank[0] == '-' || isdigit(nonblank[0])) {
            xyz_name& atom = aconf->atoms[counter];
            dbr_real x=0, y=0, z=0;
            if (extended) {
                int r = sscanf(nonblank, DBR_F " " DBR_F " " DBR_F, &x, &y, &z);
                if (r == 1) {
                    line = in.get_line(); 
                    sscanf(line, "%7s", ext_name);
                    continue;
                }
                assert(r == 3);
                assert(strlen(ext_name) > 0);
                //{
                //    mcerr << "Error in line " << i+2 << endl;
                //    dbr_abort(EXIT_FAILURE);
                //}
                strcpy(atom.name, ext_name);
            }
            else {
                dbr_real mass;
                int r = sscanf(nonblank, DBR_F" %7s "DBR_F" "DBR_F" "DBR_F, 
                                         &mass, atom.name, &x, &y, &z);
                assert(r == 5);
            }
            if (dbr_f_store_relative_coords) {
                atom.xyz[0] = x;
                atom.xyz[1] = y;
                atom.xyz[2] = z;
            }
            else {
                // fold to <0,1)
                x = x - floor(x);
                y = y - floor(y);
                z = z - floor(z);
                
                atom.xyz[0] = x * H[0][0]  +  y * H[1][0]  +  z * H[2][0];
                atom.xyz[1] = x * H[0][1]  +  y * H[1][1]  +  z * H[2][1];
                atom.xyz[2] = x * H[0][2]  +  y * H[1][2]  +  z * H[2][2];
            }
            ++counter;
            if (counter == aconf->n)
                break;
        }

        // comment line
        else if (nonblank[0] == '#') {
            nonblank++;
            if (nonblank[0] == ' ')
                nonblank++;
            aconf->comments.push_back(nonblank);
        }
    }

    aconf->pbc.v00 = H[0][0];
    aconf->pbc.v01 = H[0][1];
    aconf->pbc.v02 = H[0][2];
    aconf->pbc.v10 = H[1][0];
    aconf->pbc.v11 = H[1][1];
    aconf->pbc.v12 = H[1][2];
    aconf->pbc.v20 = H[2][0];
    aconf->pbc.v21 = H[2][1];
    aconf->pbc.v22 = H[2][2];
}

void write_atoms_to_atomeye_file(dbr_aconf const& aconf, string const& filename)
{
    if (dbr_nid != 0)
        return;
    if (dbr_verbosity >= 0)
        mcerr << "Writing atoms to AtomEye extended cfg file: " 
              << filename << endl;
    ofstream f(filename.c_str(), ios::out|ios::binary);
    if (!f) {
        mcerr << "Error. Can't open file " << filename << "for writing.\n";
        return;
    }
    f << "Number of particles =" << aconf.n << endl;
    f << "# converted by debyer\n\n";
    if (!aconf.orig_filename.empty())
        f << "# original filename: " << aconf.orig_filename << "\n";
    for (vector<string>::const_iterator i = aconf.comments.begin();
            i != aconf.comments.end(); ++i)
        f << "# " << *i << "\n";
    f << "\n";
    f << "A = 1.0 Angstrom (basic length-scale)\n";
    f << "H0(1,1) = " << setprecision(12) << aconf.pbc.v00 << " A\n";
    f << "H0(1,2) = " << setprecision(12) << aconf.pbc.v01 << " A\n";
    f << "H0(1,3) = " << setprecision(12) << aconf.pbc.v02 << " A\n";
    f << "H0(2,1) = " << setprecision(12) << aconf.pbc.v10 << " A\n";
    f << "H0(2,2) = " << setprecision(12) << aconf.pbc.v11 << " A\n";
    f << "H0(2,3) = " << setprecision(12) << aconf.pbc.v12 << " A\n";
    f << "H0(3,1) = " << setprecision(12) << aconf.pbc.v20 << " A\n";
    f << "H0(3,2) = " << setprecision(12) << aconf.pbc.v21 << " A\n";
    f << "H0(3,3) = " << setprecision(12) << aconf.pbc.v22 << " A\n";

    f << ".NO_VELOCITY.\n";
    f << "entry_count = 3\n";
    double H_1[3][3]; //inverse of pbc matrix
    dbr_inverse_3x3_matrix(aconf.pbc, H_1);
    for (int i = 0; i < aconf.n; ++i) {
        xyz_name const& atom = aconf.atoms[i];
        if (i == 0 || strcmp(atom.name, aconf.atoms[i-1].name) != 0) {
            const T_PSE *pse = FindInPSE(atom.name, 0);
            f << (pse ? pse->Mass : 1.0) << "\n";
            f << atom.name << "\n";
        }
        dbr_xyz r;
        dbr_vec3_mult_mat3x3(atom.xyz, H_1, r);
        f << r[0] << " " << r[1] << " " << r[2] << "\n";
    }
    f.close();
}

//=============    DL_POLY file reading and writing    ==================


/* checks if "atom" is a shell in core-shell model */
static bool is_shell(const char *name)
{
    static const char* postfixes[] = { "-shell", "_shell", "-shel", "_shel", 
                                        "-shl", "_shl", "-sh", "_sh" };
    const int n = sizeof(postfixes) / sizeof(postfixes[0]);
    for (int i = 0; i < n; ++i) {
        int diff = strlen(name) - strlen(postfixes[i]);
        if (diff > 0 && strcmp(postfixes[i], name+diff) == 0)
            return true;
    }
    return false;
}


void read_dlpoly_config(InputData& in, dbr_aconf *aconf)
{
    const char* line = in.get_line(); // first line is a title
    line = in.get_line(); // second line - levcfg and imcon
    int levcfg, imcon;
    int r = sscanf(line, " %d %d", &levcfg, &imcon); 
    ASSERT_FORMAT(r == 2);
    //  imcon -- periodic boundary key: 
    //      imcon       meaning
    //      0   no periodic boundaries
    //      1   cubic boundary conditions
    //      2   orthorhombic boundary conditions
    //      3   parallelepiped boundary conditions
    //      4   truncated octahedral boundary conditions
    //      5   rhombic dodecahedral boundary conditions
    //      6   x-y parallelogram boundary conditions with
    //          no periodicity in the z direction
    //      7   hexagonal prism boundary conditions
    //  
    // Possible values of levcfg: 0 - only coordinates in file
    //                            1 - coordinates and velocities
    //                            2 - coordinates, velocities and forces 

    // read boundary conditions 
    if (imcon == 0)
        ; //no PBC
    else if (imcon == 1 || imcon == 2 || imcon == 3) {
        dbr_pbc &p = aconf->pbc;

        int r = sscanf(in.get_line(), " %lf %lf %lf", &p.v00, &p.v01, &p.v02);
        ASSERT_FORMAT(r == 3);

        r = sscanf(in.get_line(), " %lf %lf %lf", &p.v10, &p.v11, &p.v12);
        ASSERT_FORMAT(r == 3);

        r = sscanf(in.get_line(), " %lf %lf %lf", &p.v20, &p.v21, &p.v22);
        ASSERT_FORMAT(r == 3);
    }
    else {
        mcerr << "Error. imcon=" << imcon << " is not supported." << endl;
        dbr_abort(EXIT_FAILURE);
    }
    
    // read atoms
    size_t counter = 0;
    size_t atoms_size = 1024;
    aconf->atoms = new xyz_name[atoms_size];
    if (levcfg > 2)
        levcfg = 2;

    while (1) {
        if (!(line = in.get_line()))
            break;
        ++counter;
        if (counter == atoms_size) 
            resize_atoms(atoms_size, &aconf->atoms);
        xyz_name& atom = aconf->atoms[counter-1];

        const char* nonblank = line;
        while (isspace(*nonblank))
            ++nonblank;

        char atom_name[8];
        atom_name[7] = 0;
        strncpy(atom_name, nonblank, 7);

        char* blank = atom_name;
        while (*blank && !isspace(*blank))
            ++blank;
        *blank = 0;

        if (is_shell(atom_name)) {
            for (int i = 0; i < levcfg+1; ++i)
                line = in.get_line();
            continue;
        }
        else {
            strncpy(atom.name, atom_name, 7);
            line = in.get_line();
            ASSERT_FORMAT(line);
            int r = sscanf(line, " " DBR_F " " DBR_F " " DBR_F, 
                                 &atom.xyz[0], &atom.xyz[1], &atom.xyz[2]);
            ASSERT_FORMAT(r == 3);
            for (int i = 1; i < levcfg+1; ++i)
                line = in.get_line();
        }
    }
    aconf->n = counter;
}


map<string,int> get_symbol_map(dbr_aconf const& aconf)
{
    map<string,int> m;
    for (int i = 0; i < aconf.n; ++i) {
        map<string,int>::iterator t = m.find(aconf.atoms[i].name);
        if (t == m.end())
            m[aconf.atoms[i].name] = 1;
        else
            t->second++;
    }
    return m;
}

template<typename T1, typename T2>
vector<T1> get_sorted_keys(map<T1,T2> const& m)
{
    vector<T1> keys;
    for (typename map<T1, T2>::const_iterator i = m.begin(); i != m.end(); ++i)
        keys.push_back(i->first);
    sort(keys.begin(), keys.end());
    return keys;
}


void write_dlpoly_file(dbr_aconf const& aconf, 
                       string const& filename, bool sorted)
{
    if (dbr_nid != 0)
        return;
    if (dbr_verbosity >= 0)
        mcerr << "Writing atoms to DL_POLY CONFIG cfg file: " 
              << filename << endl;
    // printf formatting is more useful here
    FILE *f = fopen(filename.c_str(), "wb");
    if (!f) {
        mcerr << "Error. Can't open file " << filename << "for writing.\n";
        return;
    }
    fprintf(f, "# converted by debyer\n");
    int imcon = 0;
    dbr_pbc const& pbc = aconf.pbc;
    if (pbc.v01 != 0 || pbc.v02 != 0 || pbc.v10 != 0 || pbc.v12 != 0
            || pbc.v20 != 0 || pbc.v21 != 0)
        imcon = 3;
    else if (pbc.v00 != pbc.v11 || pbc.v00 != pbc.v22)
        imcon = 2;
    else if (pbc.v00 != 0)
        imcon = 1;
    int levcfg = 0; //only coordinates in file
    fprintf(f, "%10d%10d\n", levcfg, imcon);
    if (imcon > 0) {
        fprintf(f, "%20.10f%20.10f%20.10f\n", pbc.v00, pbc.v01, pbc.v02);
        fprintf(f, "%20.10f%20.10f%20.10f\n", pbc.v10, pbc.v11, pbc.v12);
        fprintf(f, "%20.10f%20.10f%20.10f\n", pbc.v20, pbc.v21, pbc.v22);
    }
    if (sorted) {
        map<string,int> m = get_symbol_map(aconf);
        vector<string> keys = get_sorted_keys(m);
        mcerr << "Atoms in " << filename << " are in order:";
        for (vector<string>::const_iterator k = keys.begin(); 
                                                        k != keys.end(); ++k) {
            mcerr << "  " << *k << ": " << m[*k];
            //it may be inefficient for many atom species
            for (int i = 0; i < aconf.n; ++i) {
                xyz_name const& a = aconf.atoms[i];
                if (strcmp(k->c_str(), a.name) == 0) {
                    fprintf(f, "%s\n", a.name);
                    fprintf(f, "%20.8f%20.8f%20.8f\n", a.xyz[0], a.xyz[1], 
                                                                a.xyz[2]);
                }
            }
        }
        mcerr << endl;
    }
    else
        for (int i = 0; i < aconf.n; ++i) {
            xyz_name const& a = aconf.atoms[i];
            fprintf(f, "%s\n", a.name);
            fprintf(f, "%20.8f%20.8f%20.8f\n", a.xyz[0], a.xyz[1], a.xyz[2]);
        }
    fclose(f);
}
 
void read_lammps_data(InputData& /*in*/, dbr_aconf* /*aconf*/) 
{
    mcerr << "Reading LAMMPS data is not supported yet.";
    dbr_abort(1);
}

void write_lammps_data(dbr_aconf const& aconf, string const& filename)
{
    if (dbr_nid != 0)
        return;
    if (aconf.pbc.v01 != 0 || aconf.pbc.v02 != 0 
          || aconf.pbc.v10 != 0 || aconf.pbc.v12 != 0
          || aconf.pbc.v20 != 0 || aconf.pbc.v21 != 0) {
        mcerr << "Sorry, LAMMPS files with non-orthorhombic PBC "
                 "are not supported\n";
        return;
    }
    if (dbr_verbosity >= 0)
        mcerr << "Writing atoms to LAMMPS data file: " << filename << endl;
    ofstream f(filename.c_str(), ios::out|ios::binary);
    if (!f) {
        mcerr << "Error. Can't open file " << filename << "for writing.\n";
        return;
    }
    f << "# converted by debyer\n\n";
    if (!aconf.orig_filename.empty())
        f << "# original filename: " << aconf.orig_filename << "\n";
    for (vector<string>::const_iterator i = aconf.comments.begin();
            i != aconf.comments.end(); ++i)
        f << "# " << *i << "\n";
    f << "\n";
    f << "#units           real #or metal\n";
    f << "#boundary        p p p\n";
    f << "#atom_style      atomic\n";
    f << "#read_data       " << filename << "\n";
    map<string,int> m = get_symbol_map(aconf);
    vector<string> keys = get_sorted_keys(m);
    for (size_t i = 0; i != keys.size(); ++i) {
        const T_PSE *pse = FindInPSE(keys[i].c_str(), 0);
        f << "#mass            " << i+1 << " "; 
        if (pse)
            f << pse->Mass;
        else
            f << "?"; 
        f << "  #" << keys[i] << endl;
    }
    f << "\n" << aconf.n << "\tatoms\n";
    f << keys.size() << " atom types #"; 
    for (size_t i = 0; i != keys.size(); ++i) {
        f << " " << keys[i];
        m[keys[i]] = i+1;
    }
    f << "\n";
    f << 0 << " " << setprecision(12) << aconf.pbc.v00 << " xlo xhi\n";
    f << 0 << " " << setprecision(12) << aconf.pbc.v11 << " ylo yhi\n";
    f << 0 << " " << setprecision(12) << aconf.pbc.v22 << " zlo zhi\n";
    f << "\nAtoms\n\n";
    for (int i = 0; i < aconf.n; ++i) {
        xyz_name const& a = aconf.atoms[i];
        f << i+1 << "\t" << m[a.name] << "\t" 
          << a.xyz[0] << "\t" << a.xyz[1] << "\t" << a.xyz[2] << "\n"; 
    }
    f.close();
}

void write_pdb(dbr_aconf const& aconf, string const& filename)
{
    if (dbr_nid != 0)
        return;
    if (dbr_verbosity >= 0)
        mcerr << "Writing atoms to PBC data file: " << filename << endl;
    ofstream f(filename.c_str(), ios::out|ios::binary);
    if (!f) {
        mcerr << "Error. Can't open file " << filename << "for writing.\n";
        return;
    }


    f << "HEADER    converted by debyer; " << aconf.n << " atoms\n";

    // "The CRYST1 record presents the unit cell parameters, space group, 
    // and Z value. If the structure was not determined by crystallographic 
    // means, CRYST1 simply defines a unit cube."
    //
    // COLUMNS       DATA TYPE      FIELD         DEFINITION
    // -------------------------------------------------------------
    //   1 -  6       Record name    "CRYST1"
    //   7 - 15       Real(9.3)      a             a (Angstroms).
    //   16 - 24      Real(9.3)      b             b (Angstroms).
    //   25 - 33      Real(9.3)      c             c (Angstroms).
    //   34 - 40      Real(7.2)      alpha         alpha (degrees).
    //   41 - 47      Real(7.2)      beta          beta (degrees).
    //   48 - 54      Real(7.2)      gamma         gamma (degrees).
    //   56 - 66      LString        sGroup        Space group.
    //   67 - 70      Integer        z             Z value.
    //
    //   here CRYST1 is used for PBC, space group is set as P1
    dbr_c_prop p = get_cell_properties(aconf.pbc);
    f << "CRYST1" << fixed;
    for (int i = 0; i < 3; ++i)
        f << setw(9) << setprecision(3) << p.lengths[i]; 
    for (int i = 0; i < 3; ++i)
        f << setw(7) << setprecision(2) << acos(p.cosines[i]) * 180 / M_PI;
    f << " P 1       " << "   1\n";

    // "The ATOM records present the atomic coordinates for standard residues."
    // COLUMNS        DATA TYPE       FIELD         DEFINITION
    // --------------------------------------------------------------
    // 1 -6     Record name    "ATOM  "
    // 7 -11    Integer        serial      Atom serial number.
    // 13-16    Atom           name        Atom name.
    // 17       Character      altLoc      Alternate location indicator.
    // 18-20    Residue name   resName     Residue name.
    // 22       Character      chainID     Chain identifier.
    // 23-26    Integer        resSeq      Residue sequence number.
    // 27       AChar          iCode       Code for insertion of residues.
    // 31-38    Real(8.3)      x           Orthogonal coordinates for X in A
    // 39-46    Real(8.3)      y           Orthogonal coordinates for Y in A
    // 47-54    Real(8.3)      z           Orthogonal coordinates for Z in A
    // 55-60    Real(6.2)      occupancy   Occupancy.
    // 61-66    Real(6.2)      tempFactor  Temperature factor.
    // 73-76    LString(4)     segID       Segment identifier, left-justified.
    // 77-78    LString(2)     element     Element symbol, right-justified.
    // 79-80    LString(2)     charge      Charge on the atom.
    for (int i = 0; i < aconf.n; ++i) {
        f << "ATOM  " << setw(5) << i+1 
            << " " << setw(3) << left << aconf.atoms[i].name 
            <<  "          1    " << right;
        for (int j = 0; j < 3; ++j)
            f  << setw(8) << setprecision(3) << fixed << aconf.atoms[i].xyz[j];
        f << "\n";
    }
    f.close();
}

// for using in C programs
int read_atoms_c(const char* filename, xyz_name **atoms, dbr_pbc *pbc,
                 int relative)
{
    dbr_f_store_relative_coords = (bool)relative;
    InputData in(filename);
    dbr_aconf aconf = read_atoms_from_file(in);
    *atoms = aconf.atoms;
    *pbc = aconf.pbc;
    return aconf.n;
}

void free_atoms(xyz_name *atoms)
{
    delete[] atoms;
}


InputData::InputData(const char* filename_)
    : line_number(0)
{
    if (!filename_ || *filename_ == 0 || strcmp(filename_, "-") == 0) {
        stream = &std::cin;
    }
    else {
        filename = filename_;
        stream = new ifstream (filename_, ios::in|ios::binary);
        if (!(*stream)) {
            if (dbr_nid == 0)
                cerr << "Can not open file: " << filename << endl;
            dbr_abort(EXIT_FAILURE);
        }
    }
    stream->read(buffer, buffer_size-1);
    buffer[stream->gcount()] = 0;
    next_line = buffer;
}

InputData::~InputData()
{
    if (stream != &std::cin) 
        delete stream;
    stream = NULL;
}

const char* InputData::get_line()
{
    const char* line = next_line;
    if (next_line) {
        char* p = next_line + 1;
        while (*p && *p != '\n')
            ++p;
        if (*p == '\n') { 
            *p = 0;
            if (*(p-1) == '\r')
                *(p-1) = 0;
            next_line = p + 1;
        }
        else { // *p == 0
            int size = p - next_line;
            memmove(buffer, next_line, size);
            line = buffer;
            bool r = stream->getline(buffer + size, buffer_size - size - 1);
            next_line = NULL;
            if (!r)
                return NULL;
        }
    }
    else {
        bool r = stream->getline(buffer, buffer_size - 1);
        if (!r)
            return NULL;
        line = buffer;
    }
    
    ++line_number;
    return line;
}


