/*  debyer -- program for calculation of diffration patterns
 *  Copyright (C) 2006-2007 Marcin Wojdyr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: fileio.h 75 2008-07-20 15:18:03Z wojdyr $
 */

#ifndef DEBYER__FILEIO_H__
#define DEBYER__FILEIO_H__

/* we have to include debyer.h to define dbr_pbc and xyz_name */
#include "debyer.h"


#ifdef __cplusplus
extern "C" {
#endif

/* ANSI C API */
int read_atoms_c(const char* filename, xyz_name **atoms, dbr_pbc *pbc,
                 int relative);
void free_atoms(xyz_name *atoms);


/**************** the rest of this file is C++ only ******************/
#ifdef __cplusplus
}


#include <fstream>
#include <string>
#include <vector>

class InputData
{
public:
    static const int buffer_size = 1024;

    // if filename == "-", std::cin is read
    InputData(const char* filename_);
    ~InputData();

    // after the first get_line() call, get_buffer() can't be called
    const char* get_buffer() { assert(line_number == 0); return buffer; }
    // returns pointer to 0-terminated array (without a new line character). 
    // The string can be changed.
    const char* get_line();
    std::string const& get_filename() const { return filename; }
    int get_line_number() const { return line_number; }
    bool eof() const { return stream->eof(); }

private:
    std::string filename;
    int line_number;
    char buffer[1024];
    char* next_line;
    std::istream *stream;
};

// atomistic configuration 
struct dbr_aconf
{
    int n; // number of atoms
    xyz_name *atoms; // coordinates of atoms
    dbr_pbc pbc; // PBC, zeroed if not applicable
    std::string orig_filename; 
    std::vector<std::string> comments;  // no newlines here
};


extern int dbr_f_store_relative_coords;

bool is_xyz_format(char const* buffer);
bool is_plain_format(char const* /*buffer*/);

void read_xyz(InputData& in, dbr_aconf *aconf);
void read_atomeye(InputData& in, dbr_aconf *aconf);
void read_dlpoly_config(InputData& in, dbr_aconf *aconf);
void read_plain(InputData& in, dbr_aconf *aconf);
void read_lammps_data(InputData& in, dbr_aconf *aconf);

/// args: number of atoms, atom coordinates, filename
void write_atoms_to_xyz_file(dbr_aconf const& aconf, 
                             std::string const& filename);
/// args: number of atoms, atom coordinates, PBC, filename
void write_atoms_to_atomeye_file(dbr_aconf const& aconf,
                                 std::string const& filename);
/// args: number of atoms, atom coordinates, PBC, filename, 
///        bool sorted: if yes, atoms will be sorted using symbol strings
void write_dlpoly_file(dbr_aconf const& aconf,
                       std::string const& filename, bool sorted);
/// args: number of atoms, atom coordinates, PBC, filename
void write_lammps_data(dbr_aconf const& aconf, std::string const& filename);

/// args: number of atoms, atom coordinates, PBC, filename
void write_pdb(dbr_aconf const& aconf, std::string const& filename);


// buffer contains beginning of the file
dbr_aconf read_atoms_from_file(InputData &in);


// r = xyz * m
inline 
void dbr_vec3_mult_pbc(dbr_xyz const xyz, dbr_pbc const& m, dbr_xyz r)
{
    //for (int j = 0; j < 3; ++j) 
    //    r[j] = xyz[0] * m[0][j] + xyz[1] * m[1][j] + xyz[2] * m[2][j];
    r[0] = xyz[0] * m.v00 + xyz[1] * m.v10 + xyz[2] * m.v20;
    r[1] = xyz[0] * m.v01 + xyz[1] * m.v11 + xyz[2] * m.v21;
    r[2] = xyz[0] * m.v02 + xyz[1] * m.v12 + xyz[2] * m.v22;
}

inline 
void dbr_vec3_mult_mat3x3(dbr_xyz const xyz, double const m[3][3], dbr_xyz r)
{
    for (int j = 0; j < 3; ++j) 
        r[j] = xyz[0] * m[0][j] + xyz[1] * m[1][j] + xyz[2] * m[2][j];
}


#endif /* __cplusplus */
#endif /* DEBYER__FILEIO_H__ */

