//  debyer -- program for calculation of diffration patterns
//  Copyright (C) 2006 Marcin Wojdyr
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  $Id: main.cc 76 2008-07-20 23:12:22Z wojdyr $

#include <iostream>
#include <cstdlib>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstring>
#include "cmdline.h"
#include "debyer.h"
#include "fileio.h"

using namespace std;

#define mcerr \
    if (dbr_nid == 0) \
        cerr 

output_kind get_output_from_args (gengetopt_args_info const& args)
{
    if (args.RDF_given) 
        return output_rdf;
    else if (args.PDF_given) 
        return output_pdf;
    else if (args.rPDF_given) 
        return output_rpdf;
    else if (args.neutron_given) 
        return output_neutron;
    else if (args.xray_given) 
        return output_xray;
    else if (args.sf_given) 
        return output_sf;
    else 
        return output_none;
}

string default_fn(gengetopt_args_info const& args, const char* ext)
{
    char const* infn = args.inputs[0];
    char const* dot = strrchr(infn, '.');
    if (strcmp(dot+1, ext) == 0)
        return infn + string(".new");
    string base = dot ? string(infn, dot) : string(infn);
    return base + "." + ext;
}


void print_irdfs_statistics(irdfs rdfs)
{
    // show i-rdf statistics 
#ifdef __STRICT_ANSI__
    typedef long t_verylong; //may be insufficient
#else
    typedef long long t_verylong;
#endif //__STRICT_ANSI__
    t_verylong sum = 0;
    for (int i = 0; i < rdfs.pair_count; ++i) {
        for (int j = 0; j < rdfs.rdf_bins; ++j)
            sum += rdfs.data[i].nn[j];
    }
    t_verylong ac = 0;
    for (int i = 0; i < rdfs.symbol_count; ++i) 
        ac += rdfs.atom_counts[i];
    if (dbr_verbosity >= 0)
        mcerr << sum << " of " << ac * (ac-1) / 2
            << " pairs included in internal PDF.\n"; 
}

// get name of file to write ID in
const char* get_id_out_fn(gengetopt_args_info const& args)
{
    if (args.id_file_given) {
        if (args.id_file_arg)
            return args.id_file_arg;
        else {
            string default_id_file = default_fn(args, "id");
            char *allocated = new char[default_id_file.size() + 1];
            strcpy(allocated, default_id_file.c_str());
            return allocated;
        }
    }
    else
        return 0;
}

irdfs calculate_id_from_datafile(InputData &in, 
                                 gengetopt_args_info const& args)
{
    // read atoms
    irdfs rdfs;
    dbr_aconf aconf = read_atoms_from_file(in);
    if (args.pbc_a_given)
        aconf.pbc.v00 = args.pbc_a_arg;
    if (args.pbc_b_given)
        aconf.pbc.v11 = args.pbc_b_arg;
    if (args.pbc_c_given)
        aconf.pbc.v22 = args.pbc_c_arg;

    // format conversion
    if (args.write_xyz_given) 
        write_atoms_to_xyz_file(aconf, args.write_xyz_arg 
                                       ? args.write_xyz_arg 
                                       : default_fn(args, "xyz"));
    if (args.write_cfg_given) 
        write_atoms_to_atomeye_file(aconf, args.write_cfg_arg
                                           ? args.write_cfg_arg
                                           : default_fn(args, "cfg"));
    if (args.write_dlpoly_given) 
        write_dlpoly_file(aconf, args.write_dlpoly_arg
                                 ? args.write_dlpoly_arg
                                 : default_fn(args, "dlpoly"), 
                          false);
    if (args.write_dlpoly_s_given) 
        write_dlpoly_file(aconf, args.write_dlpoly_s_arg
                                 ? args.write_dlpoly_s_arg
                                 : default_fn(args, "dlpoly"),
                          true);
    if (args.write_lammps_data_given) 
        write_lammps_data(aconf, args.write_lammps_data_arg
                                 ? args.write_lammps_data_arg
                                 : default_fn(args, "lammps")); 
    if (args.write_pdb_given) 
        write_pdb(aconf, args.write_pdb_arg
                         ? args.write_pdb_arg
                         : default_fn(args, "pdb")); 
    if (args.output_group_counter == 0 && !args.id_file_given) 
        dbr_abort(EXIT_SUCCESS);

    // separete different atom types (xyz_name[] -> dbr_atoms[])
    dbr_atoms *xa = 0;
    int tc = dbr_get_atoms(aconf.n, aconf.atoms, &xa, 0);
    delete [] aconf.atoms;

    if (dbr_verbosity > 0)
        mcerr << "Elapsed " << dbr_get_elapsed() <<" s. Atoms were sorted." 
            << endl;
    if (dbr_verbosity >= 0) {
        mcerr << "Atom statistics:";
        for (int i = 0; i < tc; ++i)
            mcerr << " " << xa[i].name << ":" << xa[i].count;
        mcerr << endl;
    }

    dbr_picker picker;
    picker.probab = args.sample_given ? (dbr_real)args.sample_arg / aconf.n 
                                      : 0.;
    if (args.only_x_gt_given || args.only_x_lt_given
            || args.only_y_gt_given || args.only_y_lt_given
            || args.only_z_gt_given || args.only_z_lt_given) 
        picker.cut = 1;
    else 
        picker.cut = 0;
    picker.x_min = args.only_x_gt_given ? args.only_x_gt_arg : -DBL_MAX;
    picker.x_max = args.only_x_lt_given ? args.only_x_lt_arg : +DBL_MAX;
    picker.y_min = args.only_y_gt_given ? args.only_y_gt_arg : -DBL_MAX;
    picker.y_max = args.only_y_lt_given ? args.only_y_lt_arg : +DBL_MAX;
    picker.z_min = args.only_z_gt_given ? args.only_z_gt_arg : -DBL_MAX;
    picker.z_max = args.only_z_lt_given ? args.only_z_lt_arg : +DBL_MAX;
    picker.all = (picker.cut || picker.probab != 0. ? 0 : 1);

    // if we calculate RDF or (r)PDF, quanta and cutoff don't need to be given 
    // explicitely
    bool is_direct = dbr_is_direct(get_output_from_args(args));

    dbr_real quanta = args.quanta_arg;
    if (!args.id_file_given && is_direct && args.step_given) {
        quanta = args.step_arg;
        if (dbr_verbosity > 0)
            mcerr << "Option `quanta' set to " << quanta << endl;
    }

    dbr_real cutoff = args.cutoff_given ? args.cutoff_arg : 0.; 
    if (!args.cutoff_given && is_direct && args.to_given) {
        cutoff = args.to_arg;
        if (dbr_verbosity > 0)
            mcerr << "Option `cutoff' set to " << cutoff << endl;
    }

    //calculate internal PDFs - most of computer time is spent here
    rdfs = calculate_irdfs(tc, xa, cutoff,
                           quanta, aconf.pbc, &picker, get_id_out_fn(args));
    print_irdfs_statistics(rdfs);
    return rdfs;
}


int do_benchmark1(int count)
{
    srand(12345);
    if (dbr_verbosity >= 0)
        mcerr << "Performing ID calculation benchmark with " << count 
            << " atoms." << endl;
    xyz_name* coords = new xyz_name[count];
    dbr_real a = pow(count / 0.095, 1./3.);
    for (int i = 0; i < count; ++i) {
        strcpy(coords[i].name, i % 2 ? "Si" : "C");
        for (int j = 0; j < 3; ++j)
        coords[i].xyz[j] = a * rand() / RAND_MAX;
    }
    dbr_atoms *xa = 0;
    int tc = dbr_get_atoms(count, coords, &xa, 0);
    delete [] coords;
    if (dbr_verbosity > 0)
        mcerr << "SiC pseudo-grain (cube a=" << a << ") was generated." << endl;
    dbr_pbc pbc;
    pbc.v00 = pbc.v11 = pbc.v22 = 0.;
    dbr_picker picker;
    picker.all = 1;
    picker.cut = 0;
    picker.probab = 0.;
    irdfs rdfs = calculate_irdfs(tc, xa, 0., 0.001, pbc, &picker, 0);
    print_irdfs_statistics(rdfs);
    if (dbr_verbosity > 0)
        mcerr << "Benchmark finished." << endl;
    return 0;
}

string get_output_filename(gengetopt_args_info const& args)
{
    if (args.output_given) 
        return strcmp(args.output_arg, "stdout") != 0 ? args.output_arg : "";
    output_kind kind = get_output_from_args(args);
    if (kind == output_none) 
        return "";

    const char *ext = 0;
    if (kind == output_rdf)
        ext = "rdf";
    else if (kind == output_pdf)
        ext = "g";
    else if (kind == output_rpdf)
        ext = "rpdf";
    else if (kind == output_neutron)
        ext = args.lambda_given ? "n" : "nq";
    else if (kind == output_xray)
        ext = args.lambda_given ? "x" : "xq";
    else if (kind == output_sf)
        ext = args.lambda_given ? "s" : "sq";
    else
        assert(0);
    return default_fn(args, ext);
}

int main(int argc, char **argv)
{
    int r;
    gengetopt_args_info args;
    dbr_init(&argc, &argv);
    if (cmdline_parser(argc, argv, &args) != 0)
        dbr_abort(EXIT_FAILURE);
    if (args.quiet_given)
        dbr_verbosity = -1;
    if (args.verbose_given) {
        dbr_verbosity = 1;
        mcerr << "Verbose mode turn on." << endl;
    }
    if (args.ben1_given) {
        return do_benchmark1(args.ben1_arg);
    }
    //if (args.inputs_num > 1) {
    //    mcerr << "Only one input file can be given."; 
    if (args.inputs_num != 1) {
        mcerr << "Exactly one input file should be given." << endl; 
        dbr_abort(EXIT_FAILURE);
    }
    bool output_atoms = (args.write_xyz_given || args.write_cfg_given 
            || args.write_dlpoly_given || args.write_dlpoly_s_given
            || args.write_lammps_data_given || args.write_pdb_given);
    if (args.output_group_counter == 0 && !args.id_file_given && !output_atoms){
        mcerr << "No output format specified, nothing to do." << endl; 
        dbr_abort(EXIT_SUCCESS);
    }
    if (args.inputs_num == 0) {
        mcerr << "Reading from stdin .... TODO" << endl; 
        dbr_abort(EXIT_FAILURE);
    }

    char *infn = args.inputs[0];
    if (dbr_verbosity >= 0)
        mcerr << "examining file: " << infn << endl;

    InputData in(infn);

    output_kind kind = get_output_from_args(args);
    string ofname = get_output_filename(args);

    const char *id_magic = "debyer-id ";
    irdfs rdfs;
    if (!strncmp(in.get_buffer(), id_magic, strlen(id_magic))) {
        if (dbr_verbosity >= 0)
            mcerr << "ID file detected, skipping ID calculation phase." << endl;
        if (args.pbc_a_given || args.pbc_b_given || args.pbc_c_given) {
            mcerr << "Error. PBC options can't be applied to ID file." << endl;
            dbr_abort(EXIT_FAILURE);
        }
        if (args.quanta_given) {
            mcerr << "Error. --quanta option can't be applied to ID input file."
                << endl;
            dbr_abort(EXIT_FAILURE);
        }
        if (output_atoms) {
            mcerr << "Error. Can't deduce atom coordinates from ID file." 
                << endl;
            dbr_abort(EXIT_FAILURE);
        }
        rdfs = read_irdfs_from_file(infn);
        if (!rdfs.data)
            dbr_abort(EXIT_FAILURE);
        if (args.id_file_given) {
            mcerr << "Refusing to write ID file when input is also ID file. " 
                     "Just copy the file." << endl;
            //write_irdfs_to_file(rdfs, get_id_out_fn(args));
        }
    }
    else { //calculate id
        rdfs = calculate_id_from_datafile(in, args);
    }
    if (dbr_verbosity > 0)
        mcerr << "Elapsed " << dbr_get_elapsed() <<" s. ID is ready." << endl;

    const char *type_desc = 0;
    if (kind == output_rdf)
        type_desc = "RDF";
    else if (kind == output_pdf)
        type_desc = "PDF";
    else if (kind == output_rpdf)
        type_desc = "reduced PDF";
    else if (kind == output_neutron)
        type_desc = "neutron diffraction pattern";
    else if (kind == output_xray)
        type_desc = "x-ray diffraction pattern";
    else if (kind == output_sf)
        type_desc = "structure function";
    else { // kind == output_none
        free_irdfs(&rdfs);
        dbr_finalize();
        return 0;
    }

    if (dbr_is_inverse(kind) && args.cutoff_given 
            && !(args.ro_given || rdfs.density > 0))
        if (dbr_verbosity >= 0)
            mcerr << "No number density - no cut-off correction.\n";
    if (dbr_verbosity >= 0)
        mcerr << "Writing " << type_desc << " to file " << ofname << " ... ";
    if (dbr_is_direct(kind))
        r = write_pdfkind_to_file(kind, 
                                  rdfs,
                                  args.from_given ? args.from_arg : 0., 
                                  args.to_given ? args.to_arg : 0.,
                                  args.step_given ? args.step_arg : 0.,
                                  args.ro_given ? args.ro_arg : 0.,
                                  args.weight_arg[0],
                                  ofname.c_str());
    else {
        r = write_diffraction_to_file(kind, 
                                      rdfs,
                                      args.from_given ? args.from_arg : 0., 
                                      args.to_given ? args.to_arg : 0.,
                                      args.step_given ? args.step_arg : 0.,
                                      args.lambda_given ? args.lambda_arg : 0.,
                                      args.ro_given ? args.ro_arg : 0.,
                                      args.cutoff_given ? args.cutoff_arg : 0., 
                                      ofname.c_str());
    }
    free_irdfs(&rdfs);
    if (dbr_verbosity >= 0)
        mcerr << (r == 0 ? "OK" : "Failed") << endl;
    dbr_finalize();
    cmdline_parser_free(&args);
    if (dbr_verbosity > 0)
        mcerr << "Elapsed " << dbr_get_elapsed() <<" s. Finished." << endl;
    return 0;
}

